/** ---------------------------------------------------------------------
 * @file    usb.c
 * @brief   t\tgUSB
 * @author  ߂ ߂遗€˂Ƌl(http://yaginet.seesaa.net/)
 * @date    2011-08-24
 * 
 ------------------------------------------------------------------------*/
#include <p30f3012.h>
#include <stdio.h>
#include "usb.h"

/*--------------------------------------------------
* O[oϐ`
*--------------------------------------------------*/

/* USBt[ԍ */
static USHORT usb_frame_no = 0;

/*  */
USHORT usb_time            = 0;
/* USBڑԁilow[D+:L, D-:H] => 0x2000, full[D+:H, D-:L] => 0x4000, none[D+:L, D-:L] => 0x0000j */
USHORT usb_connected_stat  = 0x0000;

/* eobt@̎ */
static UCHAR  usb_pkt_dat_body[29];     /* 2 + 25 + 2 */

static USHORT usb_bit_sof_body[39];     /*  1 + 38              */
static USHORT usb_bit_setup_body[138];  /* (1 + 36) + (1 + 100) */
static USHORT usb_bit_in_body[58];      /* (1 + 36) + (1 +  20) */
static USHORT usb_bit_rcv_body[256];    /*      20  +      236  */

/* USBpPbgioCgf[^jobt@ */
UCHAR *usb_pkt_dat = usb_pkt_dat_body;
/* USBpPbg */
USHORT usb_pkt_len;

/* USBpPbgirbgf[^jobt@ */
USHORT *usb_bit_sof   = usb_bit_sof_body;
USHORT *usb_bit_setup = usb_bit_setup_body;
USHORT *usb_bit_in    = usb_bit_in_body;

/* M */
void  (*usb_snd)(USHORT*);
void  (*usb_rcv)(USHORT);
USHORT  usb_rcv_len;
USHORT *usb_rcv_ptr = usb_bit_rcv_body;
void  (*usb_rd)(USHORT*, USHORT);
void  (*usb_wr)(USHORT*, USHORT);


/*--------------------------------------------------
* ֐`
*--------------------------------------------------*/

/** ---------------------------------------------------------------------
 * @brief usb_ctrl_trans
 *          Rg[]
 * @param len   M
 * 
 ------------------------------------------------------------------------*/
static void usb_ctrl_trans(USHORT len)
{
    /* t[ */
    usb_make_sof_pkt();
    usb_enc_nrzi(usb_bit_sof);
    usb_wait_ms(1);
    usb_snd(usb_bit_sof);

    /* ] */
    usb_wr(usb_bit_setup, 16);
    do{
        usb_rd(usb_bit_in, len);
    }while(usb_rcv_len == 0);
}

/** ---------------------------------------------------------------------
 * @brief usb_set_adr
 *          Set_Address
 ------------------------------------------------------------------------*/
static void usb_set_adr(UCHAR adr)
{
    usb_make_token_pkt(USB_PID_SETUP, 0, 0); usb_enc_nrzi(usb_bit_setup);
    usb_make_set_adr_pkt(adr);               usb_add_nrzi(usb_bit_setup);
    usb_make_token_pkt(USB_PID_IN, 0, 0);    usb_enc_nrzi(usb_bit_in);
    usb_make_handshake_pkt(USB_PID_ACK);     usb_add_nrzi(usb_bit_in);

    usb_ctrl_trans(32 - 16);
}

/** ---------------------------------------------------------------------
 * @brief usb_get_dev_dsc
 *          Get_Descriptor foCXfBXNv^
 ------------------------------------------------------------------------*/
static void usb_get_dev_dsc(UCHAR adr)
{
    usb_make_token_pkt(USB_PID_SETUP, adr, 0); usb_enc_nrzi(usb_bit_setup);
    usb_make_get_dsc_pkt(1, 18);               usb_add_nrzi(usb_bit_setup);
    usb_make_token_pkt(USB_PID_IN, adr, 0);    usb_enc_nrzi(usb_bit_in);
    usb_make_handshake_pkt(USB_PID_ACK);       usb_add_nrzi(usb_bit_in);

    usb_ctrl_trans(176 - 16);
}

/** ---------------------------------------------------------------------
 * @brief usb_set_cfg
 *          Set_Configuration
 ------------------------------------------------------------------------*/
static void usb_set_cfg(UCHAR adr)
{
    usb_make_token_pkt(USB_PID_SETUP, adr, 0); usb_enc_nrzi(usb_bit_setup);
    usb_make_set_cfg_pkt(1);                   usb_add_nrzi(usb_bit_setup);
    usb_make_token_pkt(USB_PID_IN, adr, 0);    usb_enc_nrzi(usb_bit_in);
    usb_make_handshake_pkt(USB_PID_ACK);       usb_add_nrzi(usb_bit_in);

    usb_ctrl_trans(32 - 16);
}

/** ---------------------------------------------------------------------
 * @brief usb_calc_crc
 *          CRC̎擾
 * @param buf   Ώۃf[^
 * @param len   Ώۃf[^irbgj
 * @param crc5  TRUE:CRC5 FALSE:CRC16
 * 
 * @return  CRCl
 * 
 ------------------------------------------------------------------------*/
static USHORT usb_calc_crc(UCHAR* buf, USHORT len, BOOL crc5)
{
    USHORT i;
    USHORT mask;
    USHORT poly;
    USHORT crc;
    USHORT dat;

    mask = (crc5) ? 0x001F:0xFFFF;
    poly = (crc5) ? 0x0014:0xA001;

    crc  = mask;
    for(i = 0; i < len; i++){
        if((i & 0x7) == 0){
            dat = buf[i >> 3];
        }
        if((dat ^ crc) & 0x0001){
            crc >>= 1;
            crc  ^= poly;
        }else{
            crc >>= 1;
        }
        dat >>= 1;
    }

    return  (~crc) & mask;
}

/** ---------------------------------------------------------------------
 * @brief usb_wait_ms
 *          msecPʂł̑҂
 * @param wait_time     ҂
 * 
 ------------------------------------------------------------------------*/
void usb_wait_ms(USHORT wait_time)
{
    /* ڕW */
    USHORT next_time = usb_time + wait_time;
    /* ݃[`Ă΂ڕWƂȂ܂ő҂ */
    while(next_time != usb_time);
}

/** ---------------------------------------------------------------------
 * @brief usb_disconnect_device
 *          foCXO
 ------------------------------------------------------------------------*/
void usb_disconnect_device(void)
{
    /* 500msec҂ */
    LED_ON();
    usb_wait_ms(500);
    LED_OFF();

    /* ڑiD+, D-Ljł邱Ƃ̍ĊmF */
    if((PORTC & 0x6000) == 0){
        /* ڑԂڑɐݒ */
        usb_connected_stat = 0;
    }
}

/** ---------------------------------------------------------------------
 * @brief usb_make_sof_pkt
 *          SOFpPbg쐬ilowXs[h̏ꍇCEOPpPbg̍쐬j
 ------------------------------------------------------------------------*/
void usb_make_sof_pkt(void)
{
    USHORT crc;

    /* lowXs[h̏ꍇCEOP */
    if(usb_connected_stat & 0x2000){
        usb_pkt_len = 0;
        return ;
    }
    
    usb_pkt_dat[0] = USB_SYNC_PAT;
    usb_pkt_dat[1] = USB_PID_SOF;
    usb_pkt_dat[2] = (UCHAR)(usb_frame_no & 0x00FF);
    usb_pkt_dat[3] = (UCHAR)(usb_frame_no >> 8);

    /* t[ԍ̍XV */
    if((++usb_frame_no) & 0x0800){ usb_frame_no = 0; }

    /* CRC5 */
    crc = usb_calc_crc(usb_pkt_dat + 2, 11, TRUE);
    usb_pkt_dat[3] |= (UCHAR)(crc << 3);

    /*  */
    usb_pkt_len = 4;
}

/** ---------------------------------------------------------------------
 * @brief usb_make_token_pkt
 *          g[NpPbg쐬
 * @param pid   PID             SETUP/OUT/IN
 * @param adr   AhX        7bit
 * @param ep    Gh|Cg  4bit
 * 
 ------------------------------------------------------------------------*/
void usb_make_token_pkt(UCHAR pid, UCHAR adr, UCHAR ep)
{
    USHORT crc;
    
    usb_pkt_dat[0] = USB_SYNC_PAT;
    usb_pkt_dat[1] = pid;
    usb_pkt_dat[2] = (ep << 7) | adr;
    usb_pkt_dat[3] = ep >> 1;

    /* CRC5 */
    crc = usb_calc_crc(usb_pkt_dat + 2, 11, TRUE);
    usb_pkt_dat[3] |= (UCHAR)(crc << 3);

    /*  */
    usb_pkt_len = 4;
}

/** ---------------------------------------------------------------------
 * @brief usb_make_ctrl_pkt
 *          Rg[]pPbg쐬
 * @param type  bmRequestType
 * @param req   bReuest
 * @param val   wValue
 * @param idx   wIndex
 * @param len   wLength
 * 
 ------------------------------------------------------------------------*/
void usb_make_ctrl_pkt(UCHAR type, UCHAR req, USHORT val, USHORT idx, USHORT len)
{
    USHORT crc;
    
    usb_pkt_dat[0] = USB_SYNC_PAT;
    usb_pkt_dat[1] = USB_PID_DATA0;
    usb_pkt_dat[2] = type;                  /* bmRequestType */
    usb_pkt_dat[3] = req;                   /* bReuest       */
    usb_pkt_dat[4] = (UCHAR)(val & 0x00FF); /* wValue  L     */
    usb_pkt_dat[5] = (UCHAR)(val >> 8);     /* H             */
    usb_pkt_dat[6] = (UCHAR)(idx & 0x00FF); /* wIndex  L     */
    usb_pkt_dat[7] = (UCHAR)(idx >> 8);     /* H             */
    usb_pkt_dat[8] = (UCHAR)(len & 0x00FF); /* wLength L     */
    usb_pkt_dat[9] = (UCHAR)(len >> 8);     /* H             */

    /* CRC16 */
    crc = usb_calc_crc(usb_pkt_dat + 2, 64, FALSE);
    usb_pkt_dat[10] = (UCHAR)(crc & 0x00FF);
    usb_pkt_dat[11] = (UCHAR)(crc >> 8);

    /*  */
    usb_pkt_len = 12;
}

/** ---------------------------------------------------------------------
 * @brief usb_make_handshake_pkt
 *          nhVFCNpPbg쐬
 * @param pid   PID         ACK/NAK/STALL
 * 
 ------------------------------------------------------------------------*/
void usb_make_handshake_pkt(UCHAR pid)
{
    usb_pkt_dat[0] = USB_SYNC_PAT;
    usb_pkt_dat[1] = pid;

    /*  */
    usb_pkt_len = 2;
}

/** ---------------------------------------------------------------------
 * @brief usb_enc_nrzi
 *          NRZI
 * @param buf   o͐
 * 
 ------------------------------------------------------------------------*/
void usb_enc_nrzi(USHORT* buf)
{
    USHORT i, j;
    USHORT dat;
    USHORT out;
    USHORT bit_cnt = 0;
    USHORT n       = 6;

    out  = usb_connected_stat;
    for(i = 0; i < usb_pkt_len; i++){
        dat = usb_pkt_dat[i];
        j = 8;
        do{
            if(dat & 0x01){
                /* 1̏ꍇCo͏Ԃێ */
                /* 16񑱂ꍇC0 */
                if(--n == 0){
                    dat  &= 0xFFFE;
                    dat <<= 1; 
                    j++;
                }
            }else{
                /* 0̏ꍇCo͂𔽓] */
                out ^= 0x6000;
                n    = 6;
            }
            dat >>= 1;

            /* o͂̐ݒ */
            buf[++bit_cnt] = out;
        }while(--j != 0);
    }

    /* EOP(SE0 x 2) */
    if(usb_connected_stat & 0x2000){
        /* [Xs[h */
        buf[++bit_cnt] = 0;
        buf[++bit_cnt] = 0;
        buf[++bit_cnt] = 0x2000;
    }else{
        /* tXs[h */
        buf[++bit_cnt] = 0; /* c͓]ɕt */
    }

    /* o̓rbg */
    buf[0] = bit_cnt;
}

/** ---------------------------------------------------------------------
 * @brief usb_dec_nrzi
 *          NRZI
 * @param buf       ͌
 * @param bit_cnt   ̓rbg
 * 
 ------------------------------------------------------------------------*/
void usb_dec_nrzi(USHORT* buf, USHORT bit_cnt)
{
    USHORT i   = 0;
    USHORT j   = 0;
    USHORT n   = 2;
    USHORT in;

    /* K2Aʒu̒T */
    for( ; i < bit_cnt; i++){
        if(buf[i] ^ usb_connected_stat){
            /* K */
            if(--n == 0){ break; }
        }else{
            /* J */
            n = 2;
        }
    }
    in = buf[i++];

    /* SYNCp^[ */
    usb_pkt_dat[0] = USB_SYNC_PAT;
    usb_pkt_len    = 1;
    usb_pkt_dat[1] = 0;

    /* PIDȉ */
    n = 6;  
    for( ; i < bit_cnt; i++){
        if(buf[i] ^ in){
            /* 0:OԂƈv */
            in = buf[i];
            if(n == 0){ j--; }
            n = 6;
        }else{
            /* 1:OԂƈv */
            usb_pkt_dat[usb_pkt_len] |= (1 << j);
            n--;
        }

        /* 8bitݒ肵Ύ̃oCg */
        if(++j == 8){
            usb_pkt_dat[++usb_pkt_len] = 0;
            j = 0;
        }
    }
}

/** ---------------------------------------------------------------------
 * @brief usb_connect_device
 *          foCXڑ
 ------------------------------------------------------------------------*/
void usb_connect_device(void)
{
    USHORT i;
    UCHAR  adr = 1;
    UCHAR  bDeviceClass;

    /* ڑԂ̊mF */
    if((PORTC & 0x6000) == 0){
        /* ڑiD+, D-Lĵ܂܂ł */
        return;
    }
    /* ڑԂڑ֑J */

    /* 500msec҂ */
    LED_ON();
    usb_wait_ms(100);
    LED_OFF();

    /* ڑԂ̍ĊmF */
    usb_connected_stat = PORTC & 0x6000;
    if(usb_connected_stat == 0){
        /* ڑiD+, D-Lĵ܂܂ł */
        return;
    }else if(usb_connected_stat & 0x2000){
        /* [Xs[h[D+:L, D-:H] */
        usb_snd = usb_snd_low;
        usb_rcv = usb_rcv_low;
        usb_rd  = usb_rd_low;
        usb_wr  = usb_wr_low;
        LED_ON_L();
    }else{
        /* tXs[h[D+:H, D-:L] */
        usb_snd = usb_snd_full;
        usb_rcv = usb_rcv_full;
        usb_rd  = usb_rd_full;
        usb_wr  = usb_wr_full;
        LED_ON_R();
    }
    usb_wait_ms(100);

    /* oXZbgiD+, D-LƂj */
    TRISC = 0x0000;             /* D+, D-o */
    PORTC = 0x0000;
    usb_wait_ms(10);            /* 10msec */
    LATC  = usb_connected_stat;
    TRISC = 0xFFFF;             /* D+, D- */

    /* t[ԍZbg */
    usb_frame_no = 0;

    /* oXZbgGj[V܂40msec҂ */
    for(i = 0; i < 40; i++){
        /* 1msec̑҂SOF(EOP) */
        usb_make_sof_pkt();
        usb_enc_nrzi(usb_bit_sof);
        usb_wait_ms(1);
        usb_snd(usb_bit_sof);
    }

    /*--------------------------------------------------
    * ŏɐڑꂽfoCX̐ݒ
    *--------------------------------------------------*/

    /* Set_Address */
    usb_set_adr(adr);

    /* Set_Configuration */
    usb_set_cfg(adr);

    /* Get_Descriptor foCXfBXNv^ */
    usb_get_dev_dsc(adr);
    {
        /* foCXNX̔ */
        usb_dec_nrzi(usb_rcv_ptr, usb_rcv_len);
        bDeviceClass = usb_pkt_dat[6];
    }
    if(bDeviceClass == 0){
        /* L[{[hinuj */
        goto EXIT;
    }
    /* nutL[{[h */

    /*--------------------------------------------------
    * nu̐ݒ
    *--------------------------------------------------*/

    /* Set Port Featurei|[g1̓dj */
    {
        usb_make_token_pkt(USB_PID_SETUP, adr, 0);            usb_enc_nrzi(usb_bit_setup);
        usb_make_ctrl_pkt(0x23, USB_REQ_SET_FTR, 0x08, 1, 0); usb_add_nrzi(usb_bit_setup);
        usb_make_token_pkt(USB_PID_IN, adr, 0);               usb_enc_nrzi(usb_bit_in);
        usb_make_handshake_pkt(USB_PID_ACK);                  usb_add_nrzi(usb_bit_in);

        usb_ctrl_trans(32 - 16);
    }

    /* Clear Port Featurei|[g1̐ڑXe[^XtOj */
    {
        usb_make_token_pkt(USB_PID_SETUP, adr, 0);            usb_enc_nrzi(usb_bit_setup);
        usb_make_ctrl_pkt(0x23, USB_REQ_CLR_FTR, 0x10, 1, 0); usb_add_nrzi(usb_bit_setup);
        usb_make_token_pkt(USB_PID_IN, adr, 0);               usb_enc_nrzi(usb_bit_in);
        usb_make_handshake_pkt(USB_PID_ACK);                  usb_add_nrzi(usb_bit_in);

        usb_ctrl_trans(32 - 16);
    }

    /* Set Port Featurei|[g1̃Zbgj */
    {
        usb_make_token_pkt(USB_PID_SETUP, adr, 0);            usb_enc_nrzi(usb_bit_setup);
        usb_make_ctrl_pkt(0x23, USB_REQ_SET_FTR, 0x04, 1, 0); usb_add_nrzi(usb_bit_setup);
        usb_make_token_pkt(USB_PID_IN, adr, 0);               usb_enc_nrzi(usb_bit_in);
        usb_make_handshake_pkt(USB_PID_ACK);                  usb_add_nrzi(usb_bit_in);

        usb_ctrl_trans(32 - 16);
    }

    /* Zbg҂ */
    do{
        usb_make_token_pkt(USB_PID_IN, adr, 1);               usb_enc_nrzi(usb_bit_in);
        usb_make_handshake_pkt(USB_PID_ACK);                  usb_add_nrzi(usb_bit_in);
        do{
            /* 12msecԊuŃ[h */
            for(i = 0; i < 12; i++){
                usb_make_sof_pkt();
                usb_enc_nrzi(usb_bit_sof);
                usb_wait_ms(1);
                usb_snd(usb_bit_sof);
            }
            /* C^vg] */
            usb_rd(usb_bit_in, 40 - 16);
        }while(usb_rcv_len == 0);

        /* Get Port Status */
        usb_make_token_pkt(USB_PID_SETUP, adr, 0);            usb_enc_nrzi(usb_bit_setup);
        usb_make_ctrl_pkt(0xA3, USB_REQ_GET_STS, 0x00, 1, 4); usb_add_nrzi(usb_bit_setup);
        usb_make_token_pkt(USB_PID_IN, adr, 0);               usb_enc_nrzi(usb_bit_in);
        usb_make_handshake_pkt(USB_PID_ACK);                  usb_add_nrzi(usb_bit_in);

        usb_ctrl_trans(64 - 16);

        /* PORT_ENABLEmF */
        usb_dec_nrzi(usb_rcv_ptr, usb_rcv_len);
    }while((usb_pkt_dat[2] & 0x02) == 0);

    /* Clear Port Featurei|[g1̃ZbgXe[^XtOj */
    {
        usb_make_token_pkt(USB_PID_SETUP, adr, 0);            usb_enc_nrzi(usb_bit_setup);
        usb_make_ctrl_pkt(0x23, USB_REQ_CLR_FTR, 0x14, 1, 0); usb_add_nrzi(usb_bit_setup);
        usb_make_token_pkt(USB_PID_IN, adr, 0);               usb_enc_nrzi(usb_bit_in);
        usb_make_handshake_pkt(USB_PID_ACK);                  usb_add_nrzi(usb_bit_in);

        usb_ctrl_trans(32 - 16);
    }

    /*--------------------------------------------------
    * nu̐ݒ
    *--------------------------------------------------*/

    /* Set_Address */
    usb_set_adr(++adr);

    /* Set_Configuration */
    usb_set_cfg(adr);

EXIT:
    /*--------------------------------------------------
    * Gj[V
    *--------------------------------------------------*/
    {
        /* C^vg][hp */
        usb_make_token_pkt(USB_PID_IN, adr, 1);  usb_enc_nrzi(usb_bit_in);
        usb_make_handshake_pkt(USB_PID_ACK);     usb_add_nrzi(usb_bit_in);
    }

    /* Gj[V */
    LED_OFF();
}

